# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file has functions to create a honeycomb beam mesh."""

import numpy as _np

from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.rotation import Rotation as _Rotation
from beamme.mesh_creation_functions.beam_line import (
    create_beam_mesh_line as _create_beam_mesh_line,
)
from beamme.utils.nodes import get_min_max_nodes as _get_min_max_nodes


def create_beam_mesh_honeycomb_flat(
    mesh,
    beam_class,
    material,
    width,
    n_width,
    n_height,
    *,
    n_el=1,
    closed_width=True,
    closed_height=True,
    create_couplings=True,
):
    """Add a flat honeycomb structure. The structure will be created in the x-y
    plane.

    Args
    ----
    mesh: Mesh
        Mesh that the honeycomb will be added to.
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    width: float
        Width of one honeycomb.
    n_width: int
        Number of honeycombs in x-direction.
    n_height: int
        Number of honeycombs in y-direction.
    n_el: int
        Number of elements per beam line.
    closed_width: bool
        If the last honeycombs in x-direction will be closed.
    closed_height: bool
        If the last vertical lines in y-direction will be created.
    create_couplings: bool
        If the nodes will be connected in this function.

    Return
    ----
    return_set: GeometryName
        Set with nodes on the north, south, east and west boundaries. Those
        sets only contains end nodes of lines, not the middle ones. The set
        'all' contains all nodes.
    """

    def add_line(pointa, pointb):
        """Shortcut to add line."""
        return _create_beam_mesh_line(
            mesh_honeycomb, beam_class, material, pointa, pointb, n_el=n_el
        )

    # Geometrical shortcuts.
    sin30 = _np.sin(_np.pi / 6)
    cos30 = _np.sin(2 * _np.pi / 6)
    a = width * 0.5 / cos30
    nx = _np.array([1.0, 0.0, 0.0])
    ny = _np.array([0.0, 1.0, 0.0])
    zig_zag_x = nx * width * 0.5
    zig_zag_y = ny * a * sin30 * 0.5

    # Create the honeycomb structure
    mesh_honeycomb = _Mesh()
    origin = _np.array([0, a * 0.5 * sin30, 0])
    for i_height in range(n_height + 1):
        # Start point for this zig-zag line.
        base_row = origin + (2 * zig_zag_y + a * ny) * i_height

        # If the first node is up or down of base_row.
        if i_height % 2 == 0:
            direction = 1
        else:
            direction = -1

        for i_width in range(n_width + 1):
            base_zig_zag = base_row + direction * zig_zag_y + width * i_width * nx

            # Do not add a zig-zag line on the last run (that one is only
            # for the remaining vertical lines).
            if i_width < n_width:
                add_line(
                    base_zig_zag, base_zig_zag + zig_zag_x - 2 * direction * zig_zag_y
                )
                add_line(
                    base_zig_zag + zig_zag_x - 2 * direction * zig_zag_y,
                    base_zig_zag + nx * width,
                )

            # Check where the vertical lines start.
            if i_height % 2 == 0:
                base_vert = base_zig_zag
            else:
                base_vert = base_zig_zag + zig_zag_x - 2 * direction * zig_zag_y

            # Only add vertical lines at the end if closed_width.
            if (i_width < n_width) or (direction == 1 and closed_width):
                # Check if the vertical lines at the top should be added.
                if not (i_height == n_height) or (not closed_height):
                    add_line(base_vert, base_vert + ny * a)

    # List of nodes from the honeycomb that are candidates for connections.
    honeycomb_nodes = [node for node in mesh_honeycomb.nodes if node.is_end_node]

    # Add connections for the nodes with same positions.
    if create_couplings:
        mesh_honeycomb.couple_nodes(nodes=honeycomb_nodes)

    # Get min and max nodes of the honeycomb.
    min_max_nodes = _get_min_max_nodes(honeycomb_nodes)

    # Return the geometry set.
    return_set = _GeometryName()
    return_set["north"] = min_max_nodes["y_max"]
    return_set["east"] = min_max_nodes["x_max"]
    return_set["south"] = min_max_nodes["y_min"]
    return_set["west"] = min_max_nodes["x_min"]
    return_set["all"] = _GeometrySet(mesh_honeycomb.elements)

    mesh.add(mesh_honeycomb)

    return return_set


def create_beam_mesh_honeycomb(
    mesh,
    beam_class,
    material,
    diameter,
    n_circumference,
    n_axis,
    *,
    n_el=1,
    closed_top=True,
    vertical=True,
):
    """Wrap a honeycomb structure around a cylinder. The cylinder axis will be
    the z-axis.

    Args
    ----
    mesh: Mesh
        Mesh that the honeycomb will be added to.
    beam_class: Beam
        Class that will be used to create the beam elements.
    material: Material
        Material for the beam.
    diameter: float
        Diameter of the cylinder.
    n_circumference: int
        Number of honeycombs around the diameter. If vertical is False this
        has to be an odd number.
    n_axis: int
        Number of honeycombs in axial-direction.
    n_el: int
        Number of elements per beam line.
    closed_top: bool
        If the last honeycombs in axial-direction will be closed.
    vertical: bool
        If there are vertical lines in the honeycomb or horizontal.

    Return
    ----
    return_set: GeometryName
        Set with nodes on the top and bottom boundaries. Those sets only
        contains end nodes of lines, not the middle ones. The set "all"
        contains all nodes.
    """

    # Calculate the input values for the flat honeycomb mesh.
    if vertical:
        width = diameter * _np.pi / n_circumference
        closed_width = False
        closed_height = closed_top
        rotation = _Rotation([0, 0, 1], _np.pi / 2) * _Rotation([1, 0, 0], _np.pi / 2)
        n_height = n_axis
        n_width = n_circumference
    else:
        if not n_circumference % 2 == 0:
            raise ValueError(
                "There has to be an even number of elements along the diameter in horizontal mode. "
                "Given: {}!".format(n_circumference)
            )
        H = diameter * _np.pi / n_circumference
        r = H / (1 + _np.sin(_np.pi / 6))
        width = 2 * r * _np.cos(_np.pi / 6)
        closed_width = closed_top
        closed_height = False
        rotation = _Rotation([0, 1, 0], -0.5 * _np.pi)
        n_height = n_circumference - 1
        n_width = n_axis

    # Create the flat mesh, do not create couplings, as they will be added
    # later in this function, where also the diameter nodes will be
    # connected.
    mesh_temp = _Mesh()
    honeycomb_sets = create_beam_mesh_honeycomb_flat(
        mesh_temp,
        beam_class,
        material,
        width,
        n_width,
        n_height,
        n_el=n_el,
        closed_width=closed_width,
        closed_height=closed_height,
        create_couplings=False,
    )

    # Move the mesh to the correct position.
    mesh_temp.rotate(rotation)
    mesh_temp.translate([diameter / 2, 0, 0])
    mesh_temp.wrap_around_cylinder()

    # Add connections for the nodes with same positions.
    honeycomb_nodes = [node for node in mesh_temp.nodes if node.is_end_node]
    mesh_temp.couple_nodes(nodes=honeycomb_nodes)

    # Return the geometry set'
    return_set = _GeometryName()
    return_set["all"] = honeycomb_sets["all"]
    if vertical:
        return_set["top"] = honeycomb_sets["north"]
        return_set["bottom"] = honeycomb_sets["south"]
    else:
        return_set["top"] = honeycomb_sets["east"]
        return_set["bottom"] = honeycomb_sets["west"]

    # Add to this mesh
    mesh.add_mesh(mesh_temp)

    return return_set



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to simulate create 4C input files."""

import os
import re
from pathlib import Path
from typing import Callable, Tuple

import numpy as np
import pytest

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.function import Function
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.four_c.beam_interaction_conditions import add_beam_interaction_condition
from beamme.four_c.dbc_monitor import (
    dbc_monitor_to_mesh,
    dbc_monitor_to_mesh_all_values,
)
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.four_c.function_utility import create_linear_interpolation_function
from beamme.four_c.header_functions import (
    add_result_description,
    set_beam_contact_runtime_output,
    set_beam_contact_section,
    set_header_static,
    set_runtime_output,
)
from beamme.four_c.input_file import InputFile
from beamme.four_c.locsys_condition import LocSysCondition
from beamme.four_c.material import MaterialReissner
from beamme.four_c.model_importer import import_four_c_model
from beamme.four_c.run_four_c import run_four_c
from beamme.mesh_creation_functions.applications.beam_honeycomb import (
    create_beam_mesh_honeycomb,
)
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line
from beamme.utils.nodes import check_node_by_coordinate

# We test all test cases in this file twice. Once we only run up to the first
# call of 4C and compare the created input files, this allows to run some core
# functionalities of this file even if 4C is not available. To achieve "full"
# test coverage we also run the test, where we enforce 4C to be run.
PYTEST_4C_SIMULATION_PARAMETRIZE = [
    "enforce_four_c",
    [False, pytest.param(True, marks=pytest.mark.fourc)],
]


@pytest.fixture
def run_four_c_test(tmp_path: Path) -> Callable:
    """Provides a helper function to run a 4C simulation in a temporary
    directory.

    Args:
        tmp_path: The pytest temporary directory for the current test.

    Returns:
        Callable: A function that runs a 4C simulation.
    """

    # Counter that will track the number of calls to the _run_four_c_test function
    # for each test case (it is reset for each test case).
    run_four_c_counter = 0

    def _run_four_c_test(
        input_file: InputFile,
        n_proc: int = 2,
        restart: list[int | str | None] = [None, None],
        **kwargs,
    ) -> Tuple[Path, str]:
        """Runs a 4C simulation inside a temporary test directory.

        The function asserts that ``run_four_c`` returns 0.

        Args:
            input_file: The input file to execute.
            n_proc: Number of MPI processes to launch 4C with. Defaults to 2.
            restart: A two-element list of the form ``[restart_step, restart_from]``.
                Defaults to ``[None, None]``.
            **kwargs: Additional arguments passed directly to ``InputFile.dump()``.

        Returns:
            run_dir: The directory where the simulation ran.
            run_name: The name of the 4C run.
        """

        # Since the counter is of a base type, we have to use nonlocal to modify it.
        nonlocal run_four_c_counter
        run_four_c_counter += 1
        run_name = f"four_c_simulation_test_{run_four_c_counter}"

        # Create testing directory.
        run_dir = tmp_path / run_name
        run_dir.mkdir(parents=True, exist_ok=True)

        # Create input file.
        input_file_name = os.path.join(run_dir, run_name + ".4C.yaml")
        input_file.dump(input_file_name, add_header_information=False, **kwargs)

        return_code = run_four_c(
            input_file_name,
            run_dir,
            output_name=run_name,
            n_proc=n_proc,
            restart_step=restart[0],
            restart_from=restart[1],
        )
        assert 0 == return_code

        return run_dir, run_name

    return _run_four_c_test


def create_cantilever_model(n_steps, time_step=0.5):
    """Create a simple cantilever model.

    Args
    ----
    n_steps: int
        Number of simulation steps.
    time_step: float
        Time step size.
    """

    input_file = InputFile()
    set_header_static(input_file, time_step=time_step, n_steps=n_steps)
    input_file["IO"]["OUTPUT_BIN"] = True
    input_file["IO"]["STRUCT_DISP"] = True

    mesh = Mesh()
    ft = Function(
        [
            {
                "COMPONENT": 0,
                "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "t",
            },
        ]
    )
    mesh.add(ft)

    mat = MaterialReissner(youngs_modulus=100.0, radius=0.1)
    beam_set = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, mat, [0, 0, 0], [2, 0, 0], n_el=10
    )

    return input_file, mesh, beam_set


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
@pytest.mark.parametrize("full_import", (False, True))
def test_integration_four_c_simulation_honeycomb_sphere(
    enforce_four_c,
    full_import,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """Create the same honeycomb mesh as defined in 4C/tests/input_files/beam3r
    _herm2lin3_static_point_coupling_BTSPH_contact_stent_honeycomb_stretch_r01_
    circ10.4C.yaml The honeycomb beam is in contact with a rigid sphere, the
    sphere is moved compared to the original test file, since there are some
    problems with the contact convergence.

    The sphere is imported as an existing mesh.
    """

    # Read input file with information of the sphere and simulation.
    input_file, mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            additional_identifier="import"
        ),
        convert_input_to_mesh=full_import,
    )
    # add mesh back to the input file to check if import works
    if full_import:
        input_file.add(mesh)

    # Modify the time step options.
    input_file["STRUCTURAL DYNAMIC"]["NUMSTEP"] = 5
    input_file["STRUCTURAL DYNAMIC"]["TIMESTEP"] = 0.2

    # Delete the results given in the input file.
    input_file.pop("RESULT DESCRIPTION")

    # Add result checks.
    displacement = [0.0, -8.09347204109101170, 2.89298005937795688]
    result_descriptions = []
    nodes = [268, 188, 182]
    for node in nodes:
        for i, direction in enumerate(["x", "y", "z"]):
            result_descriptions.append(
                {
                    "STRUCTURE": {
                        "DIS": "structure",
                        "NODE": node,
                        "QUANTITY": f"disp{direction}",
                        "VALUE": displacement[i],
                        "TOLERANCE": 1e-10,
                    },
                }
            )

    input_file.add({"RESULT DESCRIPTION": result_descriptions})

    # Material for the beam.
    material = MaterialReissner(youngs_modulus=2.07e2, radius=0.1, shear_correction=1.1)

    # Create the honeycomb mesh.
    mesh_honeycomb = Mesh()
    honeycomb_set = create_beam_mesh_honeycomb(
        mesh_honeycomb,
        Beam3rHerm2Line3,
        material,
        50.0,
        10,
        4,
        n_el=1,
        closed_top=False,
    )
    mesh_honeycomb.add(honeycomb_set)
    mesh_honeycomb.rotate(Rotation([0, 0, 1], 0.5 * np.pi))

    # Functions for the boundary conditions
    ft = Function(
        [
            {
                "COMPONENT": 0,
                "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "a",
            },
            {
                "VARIABLE": 0,
                "NAME": "a",
                "TYPE": "linearinterpolation",
                "NUMPOINTS": 3,
                "TIMES": [0, 0.2, 1],
                "VALUES": [0, 1, 1],
            },
        ]
    )
    mesh_honeycomb.add(ft)

    # Change the sets to lines, only for purpose of matching the test file
    honeycomb_set["bottom"].geo_type = bme.geo.line
    honeycomb_set["top"].geo_type = bme.geo.line
    mesh_honeycomb.add(
        BoundaryCondition(
            honeycomb_set["bottom"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    mesh_honeycomb.add(
        BoundaryCondition(
            honeycomb_set["top"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                "VAL": [0, 0, 5.0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, ft, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )

    # Add the mesh to the imported solid mesh.
    input_file.add(mesh_honeycomb)

    # Check the created input file
    assert_results_close(get_corresponding_reference_file_path(), input_file)

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    # Run the input file in 4C.
    run_four_c_test(input_file)


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
@pytest.mark.parametrize("full_import", (False, True))
def test_integration_four_c_simulation_beam_and_solid_tube(
    enforce_four_c,
    full_import,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """Merge a solid tube with a beam tube and simulate them together."""

    # Create the input file and read solid mesh data.
    input_file, imported_mesh = import_four_c_model(
        input_file_path=get_corresponding_reference_file_path(
            reference_file_base_name="test_other_create_cubit_input_files_tube"
        ),
        convert_input_to_mesh=full_import,
    )
    # Add mesh back to the input file to check if import works
    if full_import:
        input_file.add(imported_mesh)

    # Add options for beam_output.
    input_file.add(
        {
            "IO/RUNTIME VTK OUTPUT/BEAMS": {
                "OUTPUT_BEAMS": True,
                "DISPLACEMENT": True,
                "USE_ABSOLUTE_POSITIONS": True,
                "TRIAD_VISUALIZATIONPOINT": True,
                "STRAINS_GAUSSPOINT": True,
            }
        }
    )

    # Add functions for boundary conditions and material.
    mesh = Mesh()
    sin = Function([{"COMPONENT": 0, "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "sin(t*2*pi)"}])
    cos = Function([{"COMPONENT": 0, "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "cos(t*2*pi)"}])
    material = MaterialReissner(youngs_modulus=1e9, radius=0.25, shear_correction=0.75)
    mesh.add(sin, cos, material)

    # Add a straight beam.
    mesh.add(material)
    cantilever_set = create_beam_mesh_line(
        mesh, Beam3rHerm2Line3, material, [2, 0, -5], [2, 0, 5], n_el=3
    )

    # Add boundary conditions.
    mesh.add(
        BoundaryCondition(
            cantilever_set["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    mesh.add(
        BoundaryCondition(
            cantilever_set["end"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [3.0, 3.0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [cos, sin, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )

    # Add result checks.
    displacements = [
        [1.50796091342925, 1.31453288915877e-8, 0.0439008100184687],
        [0.921450108160878, 1.41113401669104e-15, 0.0178350143764099],
    ]

    nodes = [32, 69]
    add_result_description(input_file, displacements, nodes)

    # Call get_unique_geometry_sets to check that this does not affect the
    # mesh creation.
    mesh.get_unique_geometry_sets(link_to_nodes="all_nodes")

    # Add mesh to input file
    input_file.add(mesh)

    # Check the created input file
    assert_results_close(
        get_corresponding_reference_file_path(),
        input_file,
    )

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    # Run the input file in 4C.
    run_four_c_test(input_file)


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
def test_integration_four_c_simulation_honeycomb_variants(
    enforce_four_c,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """Create a few different honeycomb structures."""

    # Create input file.
    input_file = InputFile()

    # Set options with different syntaxes.
    input_file.add(
        {
            "PROBLEM TYPE": {
                "PROBLEMTYPE": "Structure",
            },
            "IO": {
                "OUTPUT_BIN": False,
                "STRUCT_DISP": False,
                "VERBOSITY": "Standard",
            },
        }
    )
    input_file.add(
        {
            "IO/RUNTIME VTK OUTPUT": {
                "OUTPUT_DATA_FORMAT": "binary",
                "INTERVAL_STEPS": 1,
                "EVERY_ITERATION": False,
            }
        }
    )
    input_file.add(
        {
            "STRUCTURAL DYNAMIC": {
                "LINEAR_SOLVER": 1,
                "INT_STRATEGY": "Standard",
                "DYNAMICTYPE": "Statics",
                "PREDICT": "TangDis",
                "TIMESTEP": 1.0,
                "NUMSTEP": 666,
                "MAXTIME": 10.0,
                "TOLRES": 1.0e-4,
                "TOLDISP": 1.0e-11,
                "MAXITER": 20,
            }
        }
    )
    input_file.add(
        {
            "SOLVER 1": {
                "NAME": "Structure_Solver",
                "SOLVER": "UMFPACK",
            },
        }
    )
    input_file.add(
        {
            "IO/RUNTIME VTK OUTPUT/BEAMS": {
                "OUTPUT_BEAMS": True,
                "DISPLACEMENT": True,
                "TRIAD_VISUALIZATIONPOINT": True,
                "STRAINS_GAUSSPOINT": True,
            },
        }
    )

    # Check that we can overwrite keys
    input_file["STRUCTURAL DYNAMIC"]["NUMSTEP"] = 1

    # This does not work, because we would overwrite the entire section.
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Section(s) STRUCTURAL DYNAMIC are defined in both FourCInput objects. "
            "In order to join the FourCInput objects remove the section(s) in one of them."
        ),
    ):
        input_file.add({"STRUCTURAL DYNAMIC": {"NUMSTEP": "something"}})

    # Create four meshes with different types of honeycomb structure.
    mesh = Mesh()
    material = MaterialReissner(youngs_modulus=2.07e2, radius=0.1, shear_correction=1.1)
    ft = []
    ft.append(Function([{"SYMBOLIC_FUNCTION_OF_TIME": "t"}]))
    ft.append(Function([{"SYMBOLIC_FUNCTION_OF_TIME": "t"}]))
    ft.append(Function([{"SYMBOLIC_FUNCTION_OF_TIME": "t"}]))
    ft.append(Function([{"SYMBOLIC_FUNCTION_OF_TIME": "t"}]))
    mesh.add(ft)

    counter = 0
    for vertical in [False, True]:
        for closed_top in [False, True]:
            mesh.translate(17 * np.array([1, 0, 0]))
            honeycomb_set = create_beam_mesh_honeycomb(
                mesh,
                Beam3rHerm2Line3,
                material,
                10,
                6,
                3,
                n_el=2,
                vertical=vertical,
                closed_top=closed_top,
            )
            mesh.add(
                BoundaryCondition(
                    honeycomb_set["bottom"],
                    {
                        "NUMDOF": 9,
                        "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                        "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                        "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    },
                    bc_type=bme.bc.dirichlet,
                )
            )
            mesh.add(
                BoundaryCondition(
                    honeycomb_set["top"],
                    {
                        "NUMDOF": 9,
                        "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                        "VAL": [0.0001, 0.0001, 0.0001, 0, 0, 0, 0, 0, 0],
                        "FUNCT": [
                            ft[counter],
                            ft[counter],
                            ft[counter],
                            0,
                            0,
                            0,
                            0,
                            0,
                            0,
                        ],
                    },
                    bc_type=bme.bc.neumann,
                    double_nodes=bme.double_nodes.remove,
                )
            )
            counter += 1

    # Add mesh to input file.
    input_file.add(mesh)

    # Add result checks.
    displacements = [
        [1.31917210027397980e-01, 1.99334884558314690e-01, 6.92209310957152130e-02],
        [1.32982726482608615e-01, 2.00555145810952351e-01, 6.97003431426771458e-02],
        [7.69274209663553116e-02, 1.24993734710951515e-01, 5.86799180712692867e-02],
        [6.98802675783889299e-02, 1.09892533095288236e-01, 4.83525527530398319e-02],
    ]
    nodes = [190, 470, 711, 1071]
    add_result_description(input_file, displacements, nodes)

    # Check the created input file
    assert_results_close(get_corresponding_reference_file_path(), input_file)

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    # Run the input file in 4C.
    run_four_c_test(input_file)


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
def test_integration_four_c_simulation_rotated_beam_axis(
    enforce_four_c,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """
    Create three beams that consist of two connected lines.
    - The first case uses the same nodes for the connection of the lines,
        and the nodes are equal in this case.
    - The second case uses the same nodes for the connection of the lines,
        but the nodes have a different rotation along the basis vector 1.
    - The third case uses two nodes at the connection between the lines,
        and couples them with a coupling.
    """

    # Create input file
    input_file = InputFile()

    # Create mesh
    mesh = Mesh()

    # Set header
    set_header_static(input_file, time_step=0.05, n_steps=20)

    # Define linear function over time.
    ft = Function([{"SYMBOLIC_FUNCTION_OF_TIME": "t"}])
    mesh.add(ft)

    # Set beam material.
    mat = MaterialReissner(youngs_modulus=2.07e2, radius=0.1, shear_correction=1.1)

    # Direction of the lines and the rotation between the beams.
    direction = np.array([0.5, 1, 2])
    alpha = np.pi / 27 * 7
    force_fac = 0.01

    # Create mesh.
    for i in range(3):
        mesh_line = Mesh()

        # Create the first line.
        set_1 = create_beam_mesh_line(
            mesh_line, Beam3rHerm2Line3, mat, [0, 0, 0], 1.0 * direction, n_el=3
        )

        if not i == 0:
            # In the second case rotate the line, so the triads do not
            # match any more.
            mesh_line.rotate(Rotation(direction, alpha))

        if i == 2:
            # The third line is with couplings.
            start_node = None
        else:
            start_node = set_1["end"]

        # Add the second line.
        set_2 = create_beam_mesh_line(
            mesh_line,
            Beam3rHerm2Line3,
            mat,
            1.0 * direction,
            2.0 * direction,
            n_el=3,
            start_node=start_node,
        )

        # Add boundary conditions.
        mesh_line.add(
            BoundaryCondition(
                set_1["start"],
                {
                    "NUMDOF": 9,
                    "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                    "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                },
                bc_type=bme.bc.dirichlet,
            )
        )
        mesh_line.add(
            BoundaryCondition(
                set_2["end"],
                {
                    "NUMDOF": 9,
                    "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                    "VAL": [force_fac] * 6 + [0] * 3,
                    "FUNCT": [ft] * 6 + [0] * 3,
                },
                bc_type=bme.bc.neumann,
            )
        )

        if i == 2:
            # In the third case add a coupling.
            mesh_line.couple_nodes()

        # Add the mesh to the input file.
        mesh.add(mesh_line)

        # Each time move the whole mesh.
        mesh.translate([1, 0, 0])

    # Add result checks.
    displacements = [[1.5015284845, 0.35139255451, -1.0126517891]] * 3
    nodes = [13, 26, 40]
    add_result_description(input_file, displacements, nodes)

    # Add the mesh to the input file
    input_file.add(mesh)

    # Check the created input file
    assert_results_close(get_corresponding_reference_file_path(), input_file)

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    # Run the input file in 4C.
    run_four_c_test(input_file)
    run_four_c_test(input_file, nox_xml_file="xml_name.xml")


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
@pytest.mark.parametrize("all_values", (True, False))
def test_integration_four_c_simulation_dbc_monitor_to_input(
    enforce_four_c,
    all_values,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """Common driver to simulate a cantilever beam with Dirichlet boundary
    conditions and then apply those as Neumann boundaries.

    This can be used to test the two different functions
    by selecting one of the appropriate initial run_names:

    test_cantilever_w_dbc_monitor_to_input:
        This function explicitly tests dbc_monitor_to_input.

    test_cantilever_w_dbc_monitor_to_input_all_values:
        For the application of the boundary conditions, the last values for
        the force are used. This function explicitly tests
        dbc_monitor_to_input_all_values.
    """

    # Create and run the initial simulation.
    initial_input_file, initial_mesh, mesh_beam_set = create_cantilever_model(n_steps=2)

    initial_mesh.add(
        BoundaryCondition(
            mesh_beam_set["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    initial_mesh.add(
        BoundaryCondition(
            mesh_beam_set["end"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                "VAL": [-0.2, 1.5, 1.0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                "TAG": "monitor_reaction",
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    initial_input_file.add(
        {
            "IO/MONITOR STRUCTURE DBC": {
                "INTERVAL_STEPS": 1,
                "WRITE_CONDITION_INFORMATION": True,
                "FILE_TYPE": "yaml",
            }
        }
    )
    initial_input_file.add(initial_mesh)

    # Check the input file
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="initial"),
        initial_input_file,
    )

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    # Run the simulation in 4C
    initial_run_dir, initial_run_name = run_four_c_test(initial_input_file)

    # Create and run the second simulation.
    restart_input_file, restart_mesh, mesh_beam_set = create_cantilever_model(
        n_steps=21
    )

    restart_mesh.add(
        BoundaryCondition(
            mesh_beam_set["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )
    function_nbc = Function(
        [
            {"SYMBOLIC_FUNCTION_OF_TIME": "nbc_value"},
            {
                "VARIABLE": 0,
                "NAME": "nbc_value",
                "TYPE": "linearinterpolation",
                "NUMPOINTS": 2,
                "TIMES": [1, 11],
                "VALUES": [1, 0],
            },
        ]
    )
    restart_mesh.add(function_nbc)

    if all_values:
        dbc_monitor_to_mesh_all_values(
            restart_mesh,
            initial_run_dir / f"{initial_run_name}-102_monitor_dbc.yaml",
            n_dof=9,
            time_span=[10 * 0.5, 21 * 0.5],
            functions=[function_nbc, function_nbc, function_nbc],
        )
    else:
        dbc_monitor_to_mesh(
            restart_mesh,
            initial_run_dir / f"{initial_run_name}-102_monitor_dbc.yaml",
            n_dof=9,
            function=function_nbc,
        )

    restart_input_file.add(restart_mesh)

    displacements = [
        [-4.09988307566066690e-01, 9.93075098427816383e-01, 6.62050065618549843e-01]
    ]
    nodes = [21]
    add_result_description(restart_input_file, displacements, nodes)

    # Check the input file of the restart simulation
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="restart"),
        restart_input_file,
    )

    # Run the restart simulation
    run_four_c_test(
        restart_input_file, restart=[2, f"../{initial_run_name}/{initial_run_name}"]
    )


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
def test_integration_four_c_simulation_dirichlet_boundary_to_neumann_boundary_with_all_values(
    enforce_four_c,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """First simulate a cantilever beam with Dirichlet boundary conditions and
    then apply those as Neumann boundaries.

    For the application of the boundary conditions, all values of the
    force are used.
    """

    # Define Parameters.
    n_steps = 5  # number of simulation steps
    dt = 0.1  # time step size from create_cantilever_model

    # Create and run the initial simulation.
    initial_simulation, mesh, beam_set = create_cantilever_model(n_steps, dt)

    # Add simple lienar interpolation function.
    mesh.add(
        Function(
            [
                {
                    "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "a",
                },
                {
                    "VARIABLE": 0,
                    "NAME": "a",
                    "TYPE": "linearinterpolation",
                    "NUMPOINTS": 4,
                    "TIMES": [0, dt * n_steps, 2 * dt * n_steps, 9999999999.0],
                    "VALUES": [0.0, 1.0, 0.0, 0.0],
                },
            ]
        )
    )

    # Apply displacements to all nodes.
    for _, node in enumerate(beam_set["line"].get_all_nodes()):
        # do not constraint middle nodes
        if not node.is_middle_node:
            # Set Dirichlet conditions at one end.
            if check_node_by_coordinate(node, 0, 0):
                mesh.add(
                    BoundaryCondition(
                        GeometrySet(node),
                        {
                            "NUMDOF": 9,
                            "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                            "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                            "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                        },
                        bc_type=bme.bc.dirichlet,
                    )
                )
            else:
                # Add small displacement at other end.
                mesh.add(
                    BoundaryCondition(
                        GeometrySet(node),
                        {
                            "NUMDOF": 9,
                            "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0],
                            "VAL": [
                                0,
                                0,
                                0.25 * np.sin(node.coordinates[0] * np.pi),
                                0,
                                0,
                                0,
                                0,
                                0,
                                0,
                            ],
                            "FUNCT": [0, 0, 2, 0, 0, 0, 0, 0, 0],
                            "TAG": "monitor_reaction",
                        },
                        bc_type=bme.bc.dirichlet,
                    )
                )

    initial_simulation.add(mesh)

    # Add DB-monitor header.
    initial_simulation.add(
        {
            "IO/MONITOR STRUCTURE DBC": {
                "INTERVAL_STEPS": 1,
                "WRITE_CONDITION_INFORMATION": True,
                "FILE_TYPE": "yaml",
            }
        }
    )

    # Check the input file.
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="dirichlet"),
        initial_simulation,
    )

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    # Run the simulation in 4C.
    initial_run_dir, initial_run_name = run_four_c_test(initial_simulation)

    # Create and run the second simulation.
    force_simulation, mesh, beam_set = create_cantilever_model(2 * n_steps, dt)

    mesh.add(
        BoundaryCondition(
            beam_set["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
            },
            bc_type=bme.bc.dirichlet,
        )
    )

    # Convert the Dirichlet conditions into Neuman conditions.
    for _, _, file_names in os.walk(initial_run_dir):
        for file_name in sorted(file_names):
            if "_monitor_dbc" in file_name:
                dbc_monitor_to_mesh_all_values(
                    mesh,
                    os.path.join(initial_run_dir, file_name),
                    steps=[0, n_steps + 1],
                    time_span=[0, n_steps * dt, 2 * n_steps * dt],
                    type="hat",
                    n_dof=9,
                )

    force_simulation.add(mesh)

    displacements = [[0.0, 0.0, 0.0]]
    nodes = [21]
    add_result_description(force_simulation, displacements, nodes)

    # Compare the input file of the restart simulation.
    assert_results_close(
        get_corresponding_reference_file_path(additional_identifier="neumann"),
        force_simulation,
        atol=1e-6,
    )

    # Add runtime output.
    set_runtime_output(force_simulation)
    run_four_c_test(force_simulation)


@pytest.mark.fourc
def test_integration_four_c_simulation_cantilever_convergence(
    assert_results_close, run_four_c_test
):
    """Create multiple simulations of a cantilever beam.

    This is a legacy test that used to test the simulation manager.
    """

    def create_and_run_cantilever(n_el, *, n_proc=1):
        """Create a cantilever beam for a convergence analysis."""

        input_file = InputFile()
        set_header_static(input_file, time_step=0.25, n_steps=4)
        set_runtime_output(input_file, output_energy=True)

        mesh = Mesh()
        mat = MaterialReissner(radius=0.1, youngs_modulus=10000.0)
        beam_set = create_beam_mesh_line(
            mesh, Beam3rHerm2Line3, mat, [0, 0, 0], [1, 0, 0], n_el=n_el
        )
        mesh.add(
            BoundaryCondition(
                beam_set["start"],
                {
                    "NUMDOF": 9,
                    "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0],
                    "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                },
                bc_type=bme.bc.dirichlet,
            )
        )
        fun = Function(
            [
                {"COMPONENT": 0, "SYMBOLIC_FUNCTION_OF_SPACE_TIME": "t"},
            ]
        )
        mesh.add(
            fun,
            BoundaryCondition(
                beam_set["end"],
                {
                    "NUMDOF": 9,
                    "ONOFF": [0, 0, 1, 0, 0, 0, 0, 0, 0],
                    "VAL": [0, 0, -0.5, 0, 0, 0, 0, 0, 0],
                    "FUNCT": [0, 0, fun, 0, 0, 0, 0, 0, 0],
                },
                bc_type=bme.bc.dirichlet,
            ),
        )

        input_file.add(mesh)

        initial_run_dir, initial_run_name = run_four_c_test(input_file, n_proc=n_proc)
        my_data = np.genfromtxt(
            initial_run_dir / f"{initial_run_name}_energy.csv", delimiter=","
        )
        return my_data[-1, 2]

    results = {}
    for n_el in range(1, 7, 2):
        results[str(n_el)] = create_and_run_cantilever(n_el)
    results["ref"] = create_and_run_cantilever(40, n_proc=4)

    results_ref = {
        "5": 0.335081498526998,
        "3": 0.335055487040675,
        "1": 0.33453718896204,
        "ref": 0.335085590674607,
    }
    assert_results_close(results_ref, results)


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
def test_integration_four_c_simulation_beam_to_beam_contact_example(
    enforce_four_c,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """Small test example to show how a beam contact example with beam penalty
    contact can be set up.

    The test case consists of two beams: one beam is allocated along the x-axis and
    the other beam is located along the y-axis.
    The beam along the y-axis is placed above the other beam by an additional offset in z-Directions.
    Due to prescribed displacements at the tips of beam in y-axis, the two beams get in contact around the origin.
    """

    # Define Parameters for example
    l_beam = 2
    r_beam = 0.1
    n_ele = 11
    h = 0.25

    # Set up mesh and material.
    mesh = Mesh()
    mat = MaterialReissner(radius=r_beam, youngs_modulus=1)

    # Create a beam in x-axis.
    beam_x = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [-l_beam / 2, 0, 0],
        [l_beam / 2, 0, 0],
        n_el=n_ele,
    )

    # Apply Dirichlet condition to start and end nodes of beam 1:
    for set_name in ["start", "end"]:
        mesh.add(
            BoundaryCondition(
                beam_x[set_name],
                {
                    "NUMDOF": 9,
                    "ONOFF": [1, 1, 1, 1, 0, 0, 0, 0, 0],
                    "VAL": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    "FUNCT": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                },
                bc_type=bme.bc.dirichlet,
            )
        )

    # Create a second beam in y-axis.
    beam_y = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0, -l_beam / 2, h],
        [0, l_beam / 2, h],
        n_el=n_ele,
    )

    t = [0, 1, 1000.0]
    disp_values = [0.0, -h - r_beam, -h - r_beam]

    # Create a linear interpolation function with the displacement.
    fun = create_linear_interpolation_function(t, disp_values)

    mesh.add(fun)

    # Apply Dirichlet conditions at starting and end node to displace the beam endings.
    for set_name in ["start", "end"]:
        mesh.add(
            BoundaryCondition(
                beam_y[set_name],
                {
                    "NUMDOF": 9,
                    "ONOFF": [1, 1, 1, 0, 1, 0, 0, 0, 0],
                    "VAL": [0, 0, 1, 0, 0, 0, 0, 0, 0],
                    "FUNCT": [0, 0, fun, 0, 0, 0, 0, 0, 0],
                },
                bc_type=bme.bc.dirichlet,
            )
        )

    # Create a beam to beam contact boundary condition factory.
    add_beam_interaction_condition(
        mesh,
        beam_x["line"],
        beam_y["line"],
        bme.bc.beam_to_beam_contact,
    )

    # Create the input file
    input_file = InputFile()

    # Add the standard,static header.
    set_header_static(
        input_file,
        time_step=0.05,
        n_steps=24,
        total_time=1.2,
        write_stress="2PK",
        tol_residuum=1e-5,
        tol_increment=1e-5,
    )

    # Set the parameters for beam to beam contact.
    set_beam_contact_section(
        input_file,
        btb_penalty=50,
        penalty_regularization_g0=r_beam * 0.02,
        binning_parameters={
            "binning_cutoff_radius": 5,
            "binning_bounding_box": [-3, -3, -3, 3, 3, 3],
        },
    )

    # Add the mesh to the input file.
    input_file.add(mesh)

    # Add normal runtime output.
    set_runtime_output(input_file)

    # Add special runtime output for beam interaction.
    set_beam_contact_runtime_output(input_file, every_iteration=False)

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), input_file)

    displacements = [[1.11158519615313324e-03, 0, -1.48443346935174636e-01]]
    nodes = [13]
    add_result_description(input_file, displacements, nodes, tol=1e-8)

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    run_four_c_test(input_file, n_proc=1)


@pytest.mark.parametrize(*PYTEST_4C_SIMULATION_PARAMETRIZE)
def test_integration_four_c_simulation_locsys(
    enforce_four_c,
    assert_results_close,
    get_corresponding_reference_file_path,
    run_four_c_test,
):
    """Create a star like structure made out of 3 beams to test complex locsys
    conditions.

    We first rotate the star and then apply a prescribed displacemet to
    its center.
    """

    # Define Parameters for example
    l_beam = 2
    r_beam = 0.05
    n_ele_base = 2

    # Set up mesh and material.
    mesh = Mesh()
    mat = MaterialReissner(radius=r_beam, youngs_modulus=1)

    # List of rotations to be applied on the full stystem starting from time t=0.
    final_rotation_vector = (
        0.2 * np.pi * np.array([1, 2, 3]) / np.linalg.norm([1, 2, 3])
    )
    time_values = np.linspace(0, 1, 10)
    rotations = [
        Rotation.from_rotation_vector(final_rotation_vector * time)
        for time in time_values
    ]

    # Create the beams.
    for i in range(3):
        phi = np.pi * 2.0 / 3.0 * i
        ref_position = l_beam * np.array([np.cos(phi), np.sin(phi), 0])
        beam_set = create_beam_mesh_line(
            mesh,
            Beam3rHerm2Line3,
            mat,
            [0, 0, 0],
            ref_position,
            n_el=n_ele_base
            + i,  # A variable number of elements helps to take the symmetry out of the system.
        )

        # Get the functions describing the rotation vector components.
        ref_rotation = Rotation([0, 0, 1], phi)
        rotation_vectors = np.array(
            [(rotation * ref_rotation).get_rotation_vector() for rotation in rotations]
        )
        rotation_vector_component_functions = [
            create_linear_interpolation_function(
                time_values, rotation_vectors[:, i_dir]
            )
            for i_dir in range(3)
        ]
        mesh.add(rotation_vector_component_functions)

        # Get the "displacements" for the end node for each time step, also in the locsys coordinate system.
        displacements = np.array(
            [rotation * ref_position - ref_position for rotation in rotations]
        )
        displacements_locsys = np.array(
            [
                (rotation * ref_rotation).inv() * displacement
                for rotation, displacement in zip(rotations, displacements)
            ]
        )
        displacements_locsys_functions = [
            create_linear_interpolation_function(
                time_values, displacements_locsys[:, i_dir]
            )
            for i_dir in range(3)
        ]
        mesh.add(displacements_locsys_functions)

        # Set the boundary condition and locsys on the outer end.
        mesh.add(
            LocSysCondition(
                beam_set["end"], function_array=rotation_vector_component_functions
            )
        )
        mesh.add(
            BoundaryCondition(
                beam_set["end"],
                {
                    "NUMDOF": 9,
                    "ONOFF": [0, 1, 1] + [0] * 6,
                    "VAL": [1, 1, 1] + [0] * 6,
                    "FUNCT": displacements_locsys_functions + [0] * 6,
                },
                bc_type=bme.bc.dirichlet,
            )
        )

    # Couple the nodes at the origin together.
    mesh.couple_nodes()

    # Apply the boundary condition to the last node at the origin. This DBC is applied starting from time t=1
    f_dbc = create_linear_interpolation_function([0, 1, 2], [0, 0, 1])
    mesh.add(f_dbc)
    mesh.add(LocSysCondition(beam_set["start"], rotation=rotations[-1]))
    mesh.add(
        BoundaryCondition(
            beam_set["start"],
            {
                "NUMDOF": 9,
                "ONOFF": [0, 0, 1] + [0] * 6,
                "VAL": [0, 0, 1.0] + [0] * 6,
                "FUNCT": [0, 0, f_dbc] + [0] * 6,
            },
            bc_type=bme.bc.dirichlet,
        )
    )

    # Create the input file
    input_file = InputFile()

    # Add the standard,static header.
    set_header_static(
        input_file,
        total_time=2.0,
        n_steps=20,
        tol_residuum=1e-12,
        tol_increment=1e-12,
    )
    set_runtime_output(input_file)

    # Add the mesh to the input file.
    input_file.add(mesh)

    # Add result checks
    displacements = [
        [-6.26693202358605039e-01, 8.32271999424795572e-01, -4.56166283491497349e-01]
    ]
    nodes = [5]
    add_result_description(input_file, displacements, nodes, tol=1e-10)

    # Compare with the reference solution.
    assert_results_close(get_corresponding_reference_file_path(), input_file)

    # Check if we still have to actually run 4C.
    if not enforce_four_c:
        return

    run_four_c_test(input_file, n_proc=1)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Convert a beam to a space time surface mesh."""

from typing import Callable as _Callable
from typing import Dict as _Dict
from typing import List as _List
from typing import Tuple as _Tuple
from typing import Type as _Type
from typing import Union as _Union
from typing import cast as _cast

import numpy as _np
import vtk as _vtk

from beamme.core.conf import bme as _bme
from beamme.core.coupling import Coupling as _Coupling
from beamme.core.element_volume import VolumeElement as _VolumeElement
from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.geometry_set import GeometrySetNodes as _GeometrySetNodes
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.mesh_utils import (
    get_coupled_nodes_to_master_map as _get_coupled_nodes_to_master_map,
)
from beamme.core.node import NodeCosserat as _NodeCosserat
from beamme.utils.nodes import get_nodal_coordinates as _get_nodal_coordinates


class NodeCosseratSpaceTime(_NodeCosserat):
    """A Cosserat node in space-time.

    We add the 4th dimension time as a class variable.
    """

    def __init__(self, coordinates, rotation, time, **kwargs):
        super().__init__(coordinates, rotation, **kwargs)
        self.time = time


class SpaceTimeElement(_VolumeElement):
    """A general beam space-time surface element."""

    def __init__(self, nodes, **kwargs):
        super().__init__(nodes=nodes, **kwargs)


class SpaceTimeElementQuad4(SpaceTimeElement):
    """A space-time element with 4 nodes."""

    vtk_cell_type = _vtk.vtkQuad
    vtk_topology = list(range(4))


class SpaceTimeElementQuad9(SpaceTimeElement):
    """A space-time element with 9 nodes."""

    vtk_cell_type = _vtk.vtkQuadraticQuad
    vtk_topology = list(range(9))


def beam_to_space_time(
    mesh_space_or_generator: _Union[_Mesh, _Callable[[float], _Mesh]],
    time_duration: float,
    number_of_elements_in_time: int,
    *,
    time_start: float = 0.0,
) -> _Tuple[_Mesh, _GeometryName]:
    """Convert a beam mesh to a surface space-time mesh.

    Args:
        mesh_space_or_generator:
            Either a fixed spatial Mesh object or a function that returns the
            spatial mesh for a given time. If this is a generator, the topology
            of the mesh at the initial time is chosen for all times, only the
            positions and rotations are updated.
        time_duration:
            Total time increment to be solved with the space-time mesh
        number_of_elements_in_time:
            Number of elements in time direction
        time_start:
            Starting time for the space-time mesh. Can be used to create time slaps.
    Returns:
        Tuple (space_time_mesh, return_set)
        - space_time_mesh:
            The space time mesh. Be aware that translating / rotating this mesh
            might lead to unexpected results.
        - return_set:
            The nodes sets to be returned for the space time mesh:
                "start", "end", "left", "right", "surface"
    """

    # Get the "reference" spatial mesh
    if callable(mesh_space_or_generator):
        mesh_space_reference = mesh_space_or_generator(time_start)
    else:
        mesh_space_reference = mesh_space_or_generator

    # Perform some sanity checks
    element_types = {type(element) for element in mesh_space_reference.elements}
    if not len(element_types) == 1:
        raise ValueError(
            f"Expected all elements to be of the same type, got {element_types}"
        )
    element_type = element_types.pop()

    # Calculate global mesh properties
    number_of_nodes_in_space = len(mesh_space_reference.nodes)
    number_of_elements_in_space = len(mesh_space_reference.elements)
    space_time_element_type: _Union[
        _Type[SpaceTimeElementQuad4], _Type[SpaceTimeElementQuad9]
    ]
    if len(element_type.nodes_create) == 2:
        number_of_copies_in_time = number_of_elements_in_time + 1
        time_increment_between_nodes = time_duration / number_of_elements_in_time
        space_time_element_type = SpaceTimeElementQuad4
    elif len(element_type.nodes_create) == 3:
        number_of_copies_in_time = 2 * number_of_elements_in_time + 1
        time_increment_between_nodes = time_duration / (2 * number_of_elements_in_time)
        space_time_element_type = SpaceTimeElementQuad9
    else:
        raise TypeError(f"Got unexpected element type {element_type}")

    # Number the nodes in the original mesh
    for i_node, node in enumerate(mesh_space_reference.nodes):
        node.i_global = i_node

    # Get the nodes for the final space-time mesh
    left_nodes = []
    right_nodes = []
    space_time_nodes = []
    for i_mesh_space in range(number_of_copies_in_time):
        time = time_increment_between_nodes * i_mesh_space + time_start

        if callable(mesh_space_or_generator):
            mesh_space_current_time = mesh_space_or_generator(time)
            if (not len(mesh_space_current_time.nodes) == number_of_nodes_in_space) or (
                not len(mesh_space_current_time.elements) == number_of_elements_in_space
            ):
                raise ValueError(
                    "The number of nodes and elements does not match for the generated "
                    "space time meshes."
                )
        else:
            mesh_space_current_time = mesh_space_reference

        space_time_nodes_to_add = [
            NodeCosseratSpaceTime(
                node.coordinates, node.rotation, time, arc_length=node.arc_length
            )
            for node in mesh_space_current_time.nodes
        ]
        space_time_nodes.extend(space_time_nodes_to_add)

        if i_mesh_space == 0:
            start_nodes = space_time_nodes_to_add
        elif i_mesh_space == number_of_copies_in_time - 1:
            end_nodes = space_time_nodes_to_add

        left_nodes.append(space_time_nodes_to_add[0])
        right_nodes.append(space_time_nodes_to_add[-1])

    # Create the space time elements
    space_time_elements = []
    for i_element_time in range(number_of_elements_in_time):
        for element in mesh_space_reference.elements:
            element_node_ids = [node.i_global for node in element.nodes]
            if space_time_element_type == SpaceTimeElementQuad4:
                # Create the indices for the linear element
                first_time_row_start_index = i_element_time * number_of_nodes_in_space
                second_time_row_start_index = (
                    1 + i_element_time
                ) * number_of_nodes_in_space
                element_node_indices = [
                    first_time_row_start_index + element_node_ids[0],
                    first_time_row_start_index + element_node_ids[1],
                    second_time_row_start_index + element_node_ids[1],
                    second_time_row_start_index + element_node_ids[0],
                ]
            elif space_time_element_type == SpaceTimeElementQuad9:
                # Create the indices for the quadratic element
                first_time_row_start_index = (
                    2 * i_element_time * number_of_nodes_in_space
                )
                second_time_row_start_index = (
                    2 * i_element_time + 1
                ) * number_of_nodes_in_space
                third_time_row_start_index = (
                    2 * i_element_time + 2
                ) * number_of_nodes_in_space
                element_node_indices = [
                    first_time_row_start_index + element_node_ids[0],
                    first_time_row_start_index + element_node_ids[2],
                    third_time_row_start_index + element_node_ids[2],
                    third_time_row_start_index + element_node_ids[0],
                    first_time_row_start_index + element_node_ids[1],
                    second_time_row_start_index + element_node_ids[2],
                    third_time_row_start_index + element_node_ids[1],
                    second_time_row_start_index + element_node_ids[0],
                    second_time_row_start_index + element_node_ids[1],
                ]
            else:
                raise TypeError(
                    f"Got unexpected space time element type {space_time_element_type}"
                )

            # Add the element to the mesh
            space_time_elements.append(
                space_time_element_type(
                    [space_time_nodes[i_node] for i_node in element_node_indices]
                )
            )

    # Add joints to the space time mesh
    space_time_couplings = []
    for coupling in mesh_space_reference.boundary_conditions[
        _bme.bc.point_coupling, _bme.geo.point
    ]:
        for i_mesh_space in range(number_of_copies_in_time):
            coupling_node_ids = [
                node.i_global for node in coupling.geometry_set.get_points()
            ]
            space_time_couplings.append(
                _Coupling(
                    [
                        space_time_nodes[node_id + number_of_nodes_in_space]
                        for node_id in coupling_node_ids
                    ],
                    coupling.bc_type,
                    coupling.data,
                )
            )

    # Create the new mesh and add all the mesh items
    space_time_mesh = _Mesh()
    space_time_mesh.add(space_time_nodes)
    space_time_mesh.add(space_time_elements)
    space_time_mesh.add(space_time_couplings)

    # Create the element sets
    return_set = _GeometryName()
    return_set["start"] = _GeometrySet(start_nodes)
    return_set["end"] = _GeometrySet(end_nodes)
    return_set["left"] = _GeometrySet(left_nodes)
    return_set["right"] = _GeometrySet(right_nodes)
    return_set["surface"] = _GeometrySetNodes(_bme.geo.surface, space_time_mesh.nodes)

    return space_time_mesh, return_set


def mesh_to_data_arrays(mesh: _Mesh):
    """Get the relevant data arrays from the space time mesh."""

    element_types = list(set([type(element) for element in mesh.elements]))
    if len(element_types) > 1:
        raise ValueError("Got more than a single element type, this is not supported")
    elif not (
        element_types[0] == SpaceTimeElementQuad4
        or element_types[0] == SpaceTimeElementQuad9
    ):
        raise TypeError(
            f"Expected either SpaceTimeElementQuad4 or SpaceTimeElementQuad9, got {element_types[0]}"
        )

    _, raw_unique_nodes = _get_coupled_nodes_to_master_map(mesh, assign_i_global=True)
    unique_nodes = _cast(_List[NodeCosseratSpaceTime], raw_unique_nodes)

    n_nodes = len(unique_nodes)
    n_elements = len(mesh.elements)
    n_nodes_per_element = len(mesh.elements[0].nodes)

    coordinates = _get_nodal_coordinates(unique_nodes)
    time = _np.zeros(n_nodes)
    connectivity = _np.zeros((n_elements, n_nodes_per_element), dtype=int)
    element_rotation_vectors = _np.zeros((n_elements, n_nodes_per_element, 3))

    for i_node, node in enumerate(unique_nodes):
        time[i_node] = node.time

    for i_element, element in enumerate(mesh.elements):
        for i_node, node in enumerate(element.nodes):
            connectivity[i_element, i_node] = node.i_global
            element_rotation_vectors[i_element, i_node, :] = (
                node.rotation.get_rotation_vector()
            )

    geometry_sets = mesh.get_unique_geometry_sets()
    node_sets: _Dict[str, _Dict] = {}
    for value in geometry_sets.values():
        for geometry_set in value:
            node_set_data = {
                "node_ids": _np.array(
                    [node.i_global for node in geometry_set.get_all_nodes()]
                )
            }
            if geometry_set.name is not None:
                node_set_data["name"] = geometry_set.name
            node_sets[str(len(node_sets) + 1)] = node_set_data

    return_dict = {
        "coordinates": coordinates,
        "time": time,
        "connectivity": connectivity,
        "element_rotation_vectors": element_rotation_vectors,
        "node_sets": node_sets,
    }

    nodes_have_arc_length = {node.arc_length is not None for node in mesh.nodes}
    if len(nodes_have_arc_length) > 1:
        raise ValueError(
            "Some nodes have an arc length, some don't. This is not supported."
        )
    if nodes_have_arc_length.pop():
        # The arc length is added as an "element" property, since the same
        # node can have a different arc length depending on the element
        # (similar to the rotation vectors).
        arc_length = _np.zeros((n_elements, n_nodes_per_element))
        for i_element, element in enumerate(mesh.elements):
            for i_node, node in enumerate(element.nodes):
                connectivity[i_element, i_node] = node.i_global
                arc_length[i_element, i_node] = node.arc_length
        return_dict["arc_length"] = arc_length

    return return_dict



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file implements beam elements for 4C."""

import warnings as _warnings

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.element_beam import Beam as _Beam
from beamme.core.element_beam import generate_beam_class as _generate_beam_class
from beamme.four_c.four_c_types import (
    BeamKirchhoffConstraintType as _BeamKirchhoffConstraintType,
)
from beamme.four_c.four_c_types import (
    BeamKirchhoffParametrizationType as _BeamKirchhoffParametrizationType,
)
from beamme.four_c.four_c_types import BeamType as _BeamType
from beamme.four_c.input_file_mappings import (
    INPUT_FILE_MAPPINGS as _INPUT_FILE_MAPPINGS,
)
from beamme.four_c.material import MaterialEulerBernoulli as _MaterialEulerBernoulli
from beamme.four_c.material import MaterialKirchhoff as _MaterialKirchhoff
from beamme.four_c.material import MaterialReissner as _MaterialReissner
from beamme.four_c.material import (
    MaterialReissnerElastoplastic as _MaterialReissnerElastoplastic,
)


def dump_four_c_beam_to_list(self) -> dict:
    """Return the dictionary representing this beam element in 4C.

    Args:
        self: The beam element to be dumped.
    """

    # Check the material.
    self._check_material()

    # Gather the element data for the input file.
    element_data = type(self).four_c_element_data | self.data
    element_data["MAT"] = self.material
    if type(self).four_c_triads:
        element_data["TRIADS"] = [
            item
            for i in _INPUT_FILE_MAPPINGS["n_nodes_to_node_ordering"][len(self.nodes)]
            for item in self.nodes[i].rotation.get_rotation_vector()
        ]

    return {
        "id": self.i_global + 1,
        "cell": {
            "type": _INPUT_FILE_MAPPINGS["n_nodes_to_cell_type"][len(self.nodes)],
            "connectivity": [
                self.nodes[i]
                for i in _INPUT_FILE_MAPPINGS["n_nodes_to_node_ordering"][
                    len(self.nodes)
                ]
            ],
        },
        "data": element_data,
    }


def get_four_c_reissner_beam(n_nodes: int, is_hermite_centerline: bool) -> type[_Beam]:
    """Return a Simo-Reissner beam for 4C."""

    four_c_element_data = {
        "type": _INPUT_FILE_MAPPINGS["beam_types"][_BeamType.reissner],
        "HERMITE_CENTERLINE": is_hermite_centerline,
    }
    if is_hermite_centerline:
        coupling_fix_dict = {"NUMDOF": 9, "ONOFF": [1, 1, 1, 1, 1, 1, 0, 0, 0]}
        coupling_joint_dict = {"NUMDOF": 9, "ONOFF": [1, 1, 1, 0, 0, 0, 0, 0, 0]}
    else:
        coupling_fix_dict = {"NUMDOF": 6, "ONOFF": [1, 1, 1, 1, 1, 1]}
        coupling_joint_dict = {"NUMDOF": 6, "ONOFF": [1, 1, 1, 0, 0, 0]}

    return type(
        "BeamFourCSimoReissner",
        (_generate_beam_class(n_nodes),),
        {
            "four_c_beam_type": _BeamType.reissner,
            "four_c_triads": True,
            "four_c_element_data": four_c_element_data,
            "valid_materials": [_MaterialReissner, _MaterialReissnerElastoplastic],
            "coupling_fix_dict": coupling_fix_dict,
            "coupling_joint_dict": coupling_joint_dict,
            "dump_to_list": dump_four_c_beam_to_list,
        },
    )


def get_four_c_kirchhoff_beam(
    constraint: _BeamKirchhoffConstraintType = _BeamKirchhoffConstraintType.weak,
    parametrization: _BeamKirchhoffParametrizationType = _BeamKirchhoffParametrizationType.rot,
    is_fad: bool = True,
) -> type[_Beam]:
    """Return a Kirchhoff-Love beam for 4C."""

    # Set the parameters for this beam.
    four_c_element_data = {
        "type": _INPUT_FILE_MAPPINGS["beam_types"][_BeamType.kirchhoff],
        "CONSTRAINT": constraint.name,
        "PARAMETRIZATION": parametrization.name,
        "USE_FAD": is_fad,
    }

    # Show warning when not using rotvec.
    if not parametrization == _BeamKirchhoffParametrizationType.rot:
        _warnings.warn(
            "Use tangent based parametrization with caution, especially when "
            " applying the boundary conditions and couplings."
        )

    coupling_fix_dict = {"NUMDOF": 7, "ONOFF": [1, 1, 1, 1, 1, 1, 0]}
    coupling_joint_dict = {"NUMDOF": 7, "ONOFF": [1, 1, 1, 0, 0, 0, 0]}

    return type(
        "BeamFourCKirchhoffLove",
        (_generate_beam_class(3),),
        {
            "four_c_beam_type": _BeamType.kirchhoff,
            "four_c_triads": True,
            "four_c_element_data": four_c_element_data,
            "valid_materials": [_MaterialKirchhoff],
            "coupling_fix_dict": coupling_fix_dict,
            "coupling_joint_dict": coupling_joint_dict,
            "dump_to_list": dump_four_c_beam_to_list,
        },
    )


class BeamFourCEulerBernoulli(_generate_beam_class(2)):  # type: ignore[misc]
    """Represents a Euler Bernoulli beam element."""

    four_c_beam_type = _BeamType.euler_bernoulli
    four_c_triads = False
    four_c_element_data = {
        "type": _INPUT_FILE_MAPPINGS["beam_types"][_BeamType.euler_bernoulli]
    }

    valid_materials = [_MaterialEulerBernoulli]

    def dump_to_list(self):
        """Return a list with the (single) item representing this element."""

        # Check the material.
        self._check_material()

        # The two rotations must be the same and the x1 vector must point from
        # the start point to the end point.
        if not self.nodes[0].rotation == self.nodes[1].rotation:
            raise ValueError(
                "The two nodal rotations in Euler Bernoulli beams must be the same, i.e. the beam "
                "has to be straight!"
            )
        direction = self.nodes[1].coordinates - self.nodes[0].coordinates
        t1 = self.nodes[0].rotation * [1, 0, 0]
        if _np.linalg.norm(direction / _np.linalg.norm(direction) - t1) >= _bme.eps_pos:
            raise ValueError(
                "The rotations do not match the direction of the Euler Bernoulli beam!"
            )

        return dump_four_c_beam_to_list(self)


def get_four_c_beam(
    beam_type: _BeamType,
    *,
    n_nodes: int | None = None,
    is_hermite_centerline: bool | None = None,
    **kwargs,
) -> type[_Beam]:
    """Return an object that can be used to create beams with 4C."""

    def _check_arguments(name, value, expected):
        """Check that if an argument is given, it has the expected value."""
        if value is not None and not value == expected:
            raise ValueError(
                f"Parameter {name} with the value {value} does not match the expected value {expected}"
            )

    match beam_type:
        case _BeamType.reissner:
            # Set default values for centerline interpolation
            if n_nodes is None:
                n_nodes = 3
            if is_hermite_centerline is None:
                is_hermite_centerline = True
            return get_four_c_reissner_beam(
                n_nodes=n_nodes, is_hermite_centerline=is_hermite_centerline, **kwargs
            )
        case _BeamType.kirchhoff:
            _check_arguments("n_nodes", n_nodes, 3)
            _check_arguments("is_hermite_centerline", is_hermite_centerline, True)
            return get_four_c_kirchhoff_beam(**kwargs)
        case _BeamType.euler_bernoulli:
            _check_arguments("n_nodes", n_nodes, 2)
            _check_arguments("is_hermite_centerline", is_hermite_centerline, True)
            return BeamFourCEulerBernoulli
        case _:
            raise ValueError("Got unexpected beam type.")


# Provide shortcuts for backwards compatibility
Beam3rHerm2Line3 = get_four_c_beam(
    _BeamType.reissner, n_nodes=3, is_hermite_centerline=True
)
Beam3rLine2Line2 = get_four_c_beam(
    _BeamType.reissner, n_nodes=2, is_hermite_centerline=False
)
Beam3eb = get_four_c_beam(_BeamType.euler_bernoulli)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test that the input files created with CubitPy are up to date."""

import pytest

from tests.create_test_models import (
    create_block,
    create_single_solid_element_brick,
    create_solid_shell_meshes,
    create_tube,
)


@pytest.mark.cubitpy
def test_other_create_cubit_input_files_tube(
    tmp_path,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that the solid tube reference file is up to date."""

    result_path = tmp_path / get_corresponding_reference_file_path().name
    create_tube(result_path)
    assert_results_close(result_path, get_corresponding_reference_file_path())


@pytest.mark.cubitpy
def test_other_create_cubit_input_files_block(
    tmp_path,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that the solid block reference file is up to date."""

    result_path = tmp_path / get_corresponding_reference_file_path().name
    create_block(result_path)
    assert_results_close(result_path, get_corresponding_reference_file_path())


@pytest.mark.cubitpy
def test_other_create_cubit_input_files_solid_shell(
    tmp_path,
    get_corresponding_reference_file_path,
    assert_results_close,
):
    """Test that the solid shell reference files are up to date."""

    reference_path_blocks = get_corresponding_reference_file_path(
        additional_identifier="blocks"
    )
    reference_path_dome = get_corresponding_reference_file_path(
        additional_identifier="dome"
    )
    result_path_blocks = tmp_path / reference_path_blocks.name
    result_path_dome = tmp_path / reference_path_dome.name

    create_solid_shell_meshes(result_path_blocks, result_path_dome)
    assert_results_close(result_path_blocks, reference_path_blocks)
    assert_results_close(result_path_dome, reference_path_dome)


@pytest.mark.cubitpy
def test_other_create_cubit_input_files_single_solid_element_brick(
    tmp_path,
    get_corresponding_reference_file_path,
    get_default_test_solid_material,
    assert_results_close,
):
    """Test that the single solid element brick reference files are up to
    date."""

    reference_file = get_corresponding_reference_file_path()
    result_path = tmp_path / reference_file.name

    create_single_solid_element_brick(result_path, get_default_test_solid_material)

    assert_results_close(result_path, reference_file)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests for the beam interaction condition functionality."""

from beamme.core.boundary_condition import BoundaryCondition
from beamme.core.conf import bme
from beamme.core.geometry_set import GeometrySet
from beamme.core.mesh import Mesh
from beamme.four_c.beam_interaction_conditions import add_beam_interaction_condition
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_beamme_four_c_beam_interaction_conditions_condition_id(
    get_default_test_beam_material,
):
    """Ensure that the contact-boundary conditions ids are estimated
    correctly."""

    # Create the mesh.
    mesh = Mesh()

    # Create Material.
    mat = get_default_test_beam_material(material_type="reissner")

    # Create a beam in x-axis.
    beam_x = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0, 0, 0],
        [2, 0, 0],
        n_el=3,
    )

    # Create a second beam in y-axis.
    beam_y = create_beam_mesh_line(
        mesh,
        Beam3rHerm2Line3,
        mat,
        [0, 0, 0],
        [0, 2, 0],
        n_el=3,
    )

    # Add two contact node sets.
    id = add_beam_interaction_condition(
        mesh, beam_x["line"], beam_y["line"], bme.bc.beam_to_beam_contact
    )
    assert id == 0

    # Check if we can add the same set twice.
    id = add_beam_interaction_condition(
        mesh, beam_x["line"], beam_x["line"], bme.bc.beam_to_beam_contact
    )
    assert id == 1

    # Add some more functions to ensure that everything works as expected:
    for node in mesh.nodes:
        mesh.add(
            BoundaryCondition(
                GeometrySet(node),
                "",
                bc_type=bme.bc.dirichlet,
            )
        )

    # Add condition with higher id.
    id = add_beam_interaction_condition(
        mesh, beam_x["line"], beam_x["line"], bme.bc.beam_to_beam_contact, id=3
    )
    assert id == 3

    # Check if the id gap is filled automatically.
    id = add_beam_interaction_condition(
        mesh, beam_x["line"], beam_y["line"], bme.bc.beam_to_beam_contact
    )
    assert id == 2



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file contains utility functions for the first example."""

import numpy as np
import pyvista as pv

from beamme.utils.environment import is_testing

from .general_utils import reset_print_out


def print_matrix(name, matrix):
    """Print a matrix to the console."""
    print(f"{name}:\n{np.round(matrix, decimals=10)}")


def print_rotation_matrix(name, rotation):
    """Print the rotation matrix to the console."""
    print_matrix(name, rotation.get_rotation_matrix())


def add_cube_plot(plotter, row, col, rotation, text, *, plot_outlines=True):
    """Add a cube to the plotter."""

    plotter.subplot(row, col)

    # Define and optionally plot the original cube
    cube_original = pv.Cube()
    cube_original = cube_original.scale([3, 2, 1])
    if plot_outlines:
        plotter.add_mesh(
            cube_original, scalars=None, show_edges=True, style="wireframe"
        )

    # Rotate the cube
    rotation_vector = rotation.get_rotation_vector()
    cube = cube_original.rotate_vector(
        vector=rotation_vector, angle=np.linalg.norm(rotation_vector) * 180 / np.pi
    )

    # Define colors for each face of the cube (Rubik's cube style)
    face_colors = np.array(
        [
            [255, 165, 0],
            [255, 0, 0],
            [0, 255, 0],
            [0, 0, 255],
            [255, 255, 0],
            [255, 255, 255],
        ]
    )
    cube.cell_data["colors"] = face_colors

    # Plot the cube and the title text
    plotter.add_mesh(cube, scalars="colors", show_edges=True, rgb=True)
    plotter.add_text(text, font_size=12)

    # Plot the axes
    for i_dir, (color, name) in enumerate(
        zip(["red", "green", "blue"], ["x", "y", "z"])
    ):
        direction = np.zeros(3)
        direction[i_dir] = 3.0
        factor = 0.5
        arrow = pv.Arrow(
            start=[0, 0, 0],
            direction=direction,
            scale="auto",
            tip_radius=0.1 * factor,
            shaft_radius=0.05 * factor,
            tip_length=0.25 * factor,
        )
        plotter.add_mesh(arrow, color=color)
        plotter.add_point_labels(
            [1.075 * direction],
            [name],
            shadow=True,
            shape=None,
            show_points=False,
            font_size=32,
            justification_horizontal="center",
            justification_vertical="center",
        )

    plotter.show_axes()


class PyVistaPlotter:
    """Class to handle pyvista plotters in the examples."""

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __enter__(self):
        """Return the plotter with the given arguments."""

        self.plotter = pv.Plotter(*self.args, **self.kwargs)
        return self.plotter

    def __exit__(self, exc_type, exc_value, traceback):
        """When exiting the with statement, call this function.

        We show the plotter (except during testing and we reset the
        console print out).
        """
        if not is_testing():
            self.plotter.show()
        reset_print_out()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file includes functions to ease the creation of input files using beam
interaction potentials."""

import numpy as _np

from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.function import Function as _Function
from beamme.core.geometry_set import GeometrySet as _GeometrySet


class BeamPotential:
    """Class which provides functions for the usage of beam to beam potential
    interactions within 4C based on a potential law in form of a power law."""

    def __init__(
        self,
        *,
        pot_law_prefactor: float | int | list | _np.ndarray,
        pot_law_exponent: float | int | list | _np.ndarray,
        pot_law_line_charge_density: float | int | list | _np.ndarray,
        pot_law_line_charge_density_funcs: _Function | list | _np.ndarray | None,
    ):
        """Initialize object to enable beam potential interactions.

        Args:
            pot_law_prefactors:
                Prefactors of a potential law in form of a power law. Same number
                of prefactors and exponents/line charge densities/functions must be
                provided!
            pot_law_exponent:
                Exponents of a potential law in form of a power law. Same number
                of exponents and prefactors/line charge densities/functions must be
                provided!
            pot_law_line_charge_density:
                Line charge densities of a potential law in form of a power law.
                Same number of line charge densities and prefactors/exponents/functions
                must be provided!
            pot_law_line_charge_density_funcs:
                Functions for line charge densities of a potential law in form of a
                power law. Same number of functions and prefactors/exponents/line
                charge densities must be provided!
        """

        # if only one potential law prefactor/exponent is present, convert it
        # into a list for simplified usage
        if isinstance(pot_law_prefactor, (float, int)):
            pot_law_prefactor = [pot_law_prefactor]
        if isinstance(pot_law_exponent, (float, int)):
            pot_law_exponent = [pot_law_exponent]
        if isinstance(pot_law_line_charge_density, (float, int)):
            pot_law_line_charge_density = [pot_law_line_charge_density]
        if (
            isinstance(pot_law_line_charge_density_funcs, _Function)
            or pot_law_line_charge_density_funcs is None
        ):
            pot_law_line_charge_density_funcs = [pot_law_line_charge_density_funcs]

        # check if same number of prefactors and exponents are provided
        if (
            not len(pot_law_prefactor)
            == len(pot_law_exponent)
            == len(pot_law_line_charge_density)
        ):
            raise ValueError(
                "Number of potential law prefactors do not match potential law exponents or potential line charge density!"
            )

        self.pot_law_prefactor = pot_law_prefactor
        self.pot_law_exponent = pot_law_exponent
        self.pot_law_line_charge_density = pot_law_line_charge_density
        self.pot_law_line_charge_density_funcs = pot_law_line_charge_density_funcs

    def create_header(
        self,
        *,
        potential_type: str,
        evaluation_strategy: str,
        cutoff_radius: float,
        regularization_type: str | None = None,
        regularization_separation: float,
        integration_segments: int,
        gauss_points: int,
        potential_reduction_length: float | int | None = None,
        automatic_differentiation: bool,
        choice_master_slave: str | None,
        runtime_output_interval_steps: int | None = None,
        runtime_output_every_iteration: bool,
        runtime_output_force: bool,
        runtime_output_moment: bool,
        runtime_output_uids: bool,
        runtime_output_per_ele_pair: bool,
    ) -> dict:
        """Set the basic header options for beam potential interactions.

        Args:
            potential_type:
                Type of applied potential.
            evaluation_strategy:
                Strategy to evaluate interaction potential.
            cutoff_radius:
                Neglect all contributions at separation larger than this cutoff
                radius.
            regularization_type:
                Type of regularization to use for force law at separations below
                specified separation.
            regularization_separation:
                Use specified regularization type for separations smaller than
                this value.
            integration_segments:
                Number of integration segments to be used per beam element.
            gauss_points:
                Number of Gauss points to be used per integration segment.
            potential_reduction_length:
                Potential is smoothly decreased within this length when using the
                single length specific (SBIP) approach to enable an axial pull off
                force.
            automatic_differentiation:
                Use automatic differentiation via FAD.
            choice_master_slave:
                Rule how to assign the role of master and slave to beam elements (if
                applicable).

            runtime_output:
                If the output for beam potential should be written.
            runtime_output_interval_steps:
                Interval at which output is written.
            runtime_output_every_iteration:
                If output at every Newton iteration should be written.
            runtime_output_force:
                If the forces should be written.
            runtime_output_moment:
                If the moments should be written.
            runtime_output_uids:
                If the unique ids should be written.
            runtime_output_per_ele_pair:
                If the forces/moments should be written per element pair.

        Returns:
            Header for beam potential interactions.
        """

        header = {
            "beam_potential": {
                "type": potential_type,
                "strategy": evaluation_strategy,
                "potential_law_prefactors": self.pot_law_prefactor,
                "potential_law_exponents": self.pot_law_exponent,
                "automatic_differentiation": automatic_differentiation,
                "cutoff_radius": cutoff_radius,
                "n_integration_segments": integration_segments,
                "n_gauss_points": gauss_points,
                "potential_reduction_length": potential_reduction_length,
            }
        }

        if regularization_type is not None:
            header["beam_potential"]["regularization"] = {
                "type": regularization_type,
                "separation": regularization_separation,
            }

        if choice_master_slave is not None:
            header["beam_potential"]["choice_master_slave"] = choice_master_slave

        if runtime_output_interval_steps is not None:
            header["beam_potential"]["runtime_output"] = {
                "interval_steps": runtime_output_interval_steps,
                "force": runtime_output_force,
                "moment": runtime_output_moment,
                "every_iteration": runtime_output_every_iteration,
                "write_force_moment_per_elementpair": runtime_output_per_ele_pair,
                "write_uids": runtime_output_uids,
            }

        return header

    def create_potential_charge_conditions(
        self, *, geometry_set: _GeometrySet
    ) -> list[_BoundaryCondition]:
        """Create potential charge conditions.

        Args:
            geometry_set:
                Add potential charge condition to this set.

        Returns:
            List of boundary conditions for potential charge.
        """

        bcs = []

        for i, (line_charge, func) in enumerate(
            zip(
                self.pot_law_line_charge_density, self.pot_law_line_charge_density_funcs
            )
        ):
            bc = _BoundaryCondition(
                geometry_set,
                {"POTLAW": i + 1, "VAL": line_charge, "FUNCT": func},
                bc_type="DESIGN LINE BEAM POTENTIAL CHARGE CONDITIONS",
            )

            bcs.append(bc)

        return bcs



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module defines the class that is used to create an input file for
Abaqus."""

from enum import Enum as _Enum
from enum import auto as _auto

import numpy as _np

from beamme.core.conf import INPUT_FILE_HEADER as _INPUT_FILE_HEADER
from beamme.core.conf import bme as _bme
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.mesh import Mesh as _Mesh
from beamme.core.mesh_utils import (
    get_coupled_nodes_to_master_map as _get_coupled_nodes_to_master_map,
)
from beamme.core.rotation import smallest_rotation as _smallest_rotation

# Format template for different number types.
F_INT = "{:6d}"
F_FLOAT = "{: .14e}"


def set_i_global(data_list, *, start_index=0):
    """Set i_global in every item of data_list.

    Args
    ----
    data_list:
        List containing the items that should be numbered
    start_index: int
        Starting index of the numbering
    """

    # A check is performed that every entry in data_list is unique.
    if len(data_list) != len(set(data_list)):
        raise ValueError("Elements in data_list are not unique!")

    # Set the values for i_global.
    for i, item in enumerate(data_list):
        item.i_global = i + start_index


def get_set_lines(set_type, items, name):
    """Get the Abaqus input file lines for a set of items (max 16 items per
    row)"""
    max_entries_per_line = 16
    lines = ["*{}, {}={}".format(set_type, set_type.lower(), name)]
    set_ids = [item.i_global + 1 for item in items]
    set_ids.sort()
    set_ids = [
        set_ids[i : i + max_entries_per_line]
        for i in range(0, len(set_ids), max_entries_per_line)
    ]
    for ids in set_ids:
        lines.append(", ".join([F_INT.format(id) for id in ids]))
    return lines


class AbaqusBeamNormalDefinition(_Enum):
    """Enum for different ways to define the beam cross-section normal.

    For more information see the Abaqus documentation on: "Beam element cross-section orientation"
    and the function `AbaqusInputFile.calculate_cross_section_normal_data`.
    """

    normal_and_extra_node = _auto()
    """Create an extra node and the nodal normal information for each node."""

    normal = _auto()
    """Create the nodal normal information for each node."""


class AbaqusInputFile(object):
    """This class represents an Abaqus input file."""

    def __init__(self, mesh: _Mesh):
        """Initialize the input file.

        Args
        ----
        mesh: Mesh()
            Mesh to be used in this input file.
        """
        self.mesh = mesh

    def write_input_file(
        self,
        file_path,
        *,
        normal_definition=AbaqusBeamNormalDefinition.normal_and_extra_node,
    ):
        """Write the ASCII input file to disk.

        Args
        ----
        file_path: path
            Path on the disk, where the input file should be stored.
        normal_definition: AbaqusBeamNormalDefinition
            How the beam cross-section should be defined.
        """

        # Write the input file to disk
        with open(file_path, "w") as input_file:
            input_file.write(self.get_input_file_string(normal_definition))
            input_file.write("\n")

    def get_input_file_string(self, normal_definition):
        """Generate the string for the Abaqus input file."""

        # Perform some checks on the mesh.
        if _bme.check_overlapping_elements:
            self.mesh.check_overlapping_elements()

        # Assign global indices to all materials
        set_i_global(self.mesh.materials)

        # Calculate the required cross-section normal data
        self.calculate_cross_section_normal_data(normal_definition)

        # Add the lines to the input file
        input_file_lines = []
        input_file_lines.extend(["** " + line for line in _INPUT_FILE_HEADER])
        input_file_lines.extend(self.get_nodes_lines())
        input_file_lines.extend(self.get_element_lines())
        input_file_lines.extend(self.get_material_lines())
        input_file_lines.extend(self.get_set_lines())
        return "\n".join(input_file_lines)

    def calculate_cross_section_normal_data(self, normal_definition):
        """Evaluate all data that is required to fully specify the cross-
        section orientation in Abaqus. The evaluated data is stored in the
        elements.

        For more information see the Abaqus documentation on: "Beam element cross-section orientation"

        Args
        ----
        normal_definition: AbaqusBeamNormalDefinition
            How the beam cross-section should be defined.
        """

        def normalize(vector):
            """Normalize a vector."""
            return vector / _np.linalg.norm(vector)

        # Reset possibly existing data stored in the elements
        # element.n1_orientation_node: list(float)
        #     The coordinates of an additional (dummy) node connected to the
        #     element to define its approximate n1 direction. It this is None,
        #     no additional node will be added to the input file.
        # element.n1_node_id: str
        #     The global ID in the input file for the additional orientation
        #     node.
        # element.n2: list(list(float)):
        #     A list containing possible explicit normal definitions for each
        #     element node. All entries that are not None will be added to the
        #     *NORMAL section of the input file.

        for element in self.mesh.elements:
            element.n1_position = None
            element.n1_node_id = None
            element.n2 = [None for i_node in range(len(element.nodes))]

        if (
            normal_definition == AbaqusBeamNormalDefinition.normal
            or normal_definition == AbaqusBeamNormalDefinition.normal_and_extra_node
        ):
            # In this case we take the beam tangent from the first to the second node
            # and calculate an ortho-normal triad based on this direction. We do this
            # via a smallest rotation mapping from the triad of the first node onto
            # the tangent.

            for element in self.mesh.elements:
                node_1 = element.nodes[0].coordinates
                node_2 = element.nodes[1].coordinates
                t = normalize(node_2 - node_1)

                rotation = element.nodes[0].rotation
                cross_section_rotation = _smallest_rotation(rotation, t)

                if (
                    normal_definition
                    == AbaqusBeamNormalDefinition.normal_and_extra_node
                ):
                    element.n1_position = node_1 + cross_section_rotation * [
                        0.0,
                        1.0,
                        0.0,
                    ]
                element.n2[0] = cross_section_rotation * [0.0, 0.0, 1.0]
        else:
            raise ValueError(f"Got unexpected normal_definition {normal_definition}")

    def get_nodes_lines(self):
        """Get the lines for the input file that represent the nodes."""

        # The nodes require postprocessing, as we have to identify coupled nodes in Abaqus.
        # Internally in Abaqus, coupled nodes are a single node with different normals for the
        # connected element. Therefore, for nodes which are coupled to each other, we keep the
        # same global ID while still keeping the individual nodes.
        _, unique_nodes = _get_coupled_nodes_to_master_map(
            self.mesh, assign_i_global=True
        )

        # Number the remaining nodes and create nodes for the input file
        input_file_lines = ["*Node"]
        for node in unique_nodes:
            input_file_lines.append(
                (", ".join([F_INT] + 3 * [F_FLOAT])).format(
                    node.i_global + 1, *node.coordinates
                )
            )

        # Check if we need to write additional nodes for the element cross-section directions
        node_counter = len(unique_nodes)
        for element in self.mesh.elements:
            if element.n1_position is not None:
                node_counter += 1
                input_file_lines.append(
                    (", ".join([F_INT] + 3 * [F_FLOAT])).format(
                        node_counter, *element.n1_position
                    )
                )
                element.n1_node_id = node_counter

        return input_file_lines

    def get_element_lines(self):
        """Get the lines for the input file that represent the elements."""

        # Sort the elements after their types.
        element_types = {}
        for element in self.mesh.elements:
            element_type = element.beam_type
            if element_type in element_types.keys():
                element_types[element_type].append(element)
            else:
                element_types[element_type] = [element]

        # Write the element connectivity.
        element_count = 0
        element_lines = []
        normal_lines = ["*Normal, type=element"]
        for element_type, elements in element_types.items():
            # Number the elements of this type
            set_i_global(elements, start_index=element_count)

            # Set the element connectivity, possibly including the n1 direction node
            element_lines.append("*Element, type={}".format(element_type))
            for element in elements:
                node_ids = [node.i_global + 1 for node in element.nodes]
                if element.n1_node_id is not None:
                    node_ids.append(element.n1_node_id)
                line_ids = [element.i_global + 1] + node_ids
                element_lines.append(", ".join(F_INT.format(i) for i in line_ids))

                # Set explicit normal definitions for the nodes
                for i_node, n2 in enumerate(element.n2):
                    if n2 is not None:
                        node = element.nodes[i_node]
                        normal_lines.append(
                            (", ".join(2 * [F_INT] + 3 * [F_FLOAT])).format(
                                element.i_global + 1, node.i_global + 1, *n2
                            )
                        )

            element_count += len(elements)

        if len(normal_lines) > 1:
            return element_lines + normal_lines
        else:
            return element_lines

    def get_material_lines(self):
        """Get the lines for the input file that represent the element sets
        with the same material."""

        materials = {}
        for element in self.mesh.elements:
            element_material = element.material
            if element_material in materials.keys():
                materials[element_material].append(element)
            else:
                materials[element_material] = [element]

        # Create the element sets for the different materials.
        input_file_lines = []
        for material, elements in materials.items():
            material_name = material.dump_to_list()[0]
            input_file_lines.extend(get_set_lines("Elset", elements, material_name))
        return input_file_lines

    def get_set_lines(self):
        """Add lines to the input file that represent node and element sets."""

        input_file_lines = []
        for point_set in self.mesh.geometry_sets[_bme.geo.point]:
            if point_set.name is None:
                raise ValueError("Sets added to the mesh have to have a valid name!")
            input_file_lines.extend(
                get_set_lines("Nset", point_set.get_points(), point_set.name)
            )
        for line_set in self.mesh.geometry_sets[_bme.geo.line]:
            if line_set.name is None:
                raise ValueError("Sets added to the mesh have to have a valid name!")
            if isinstance(line_set, _GeometrySet):
                input_file_lines.extend(
                    get_set_lines(
                        "Elset", line_set.geometry_objects[_bme.geo.line], line_set.name
                    )
                )
            else:
                raise ValueError(
                    "Line sets can only be exported to Abaqus if they are defined with the beam elements"
                )
        return input_file_lines



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test the NURBS mesh creation functions."""

import numpy as np
import splinepy

from beamme.core.mesh import Mesh
from beamme.core.rotation import Rotation
from beamme.mesh_creation_functions.nurbs_generic import (
    add_geomdl_nurbs_to_mesh,
    add_splinepy_nurbs_to_mesh,
)
from beamme.mesh_creation_functions.nurbs_geometries import (
    create_nurbs_brick,
    create_nurbs_cylindrical_shell_sector,
    create_nurbs_flat_plate_2d,
    create_nurbs_hemisphere_surface,
    create_nurbs_hollow_cylinder_segment_2d,
    create_nurbs_sphere_surface,
    create_nurbs_torus_surface,
)
from beamme.mesh_creation_functions.nurbs_utils import (
    ensure_3d_splinepy_object,
    translate_splinepy,
)


def test_integration_mesh_creation_functions_nurbs_hollow_cylinder_segment_2d(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a two dimensional hollow cylinder segment."""

    # Create the surface of a quarter of a hollow cylinder
    surf_obj = create_nurbs_hollow_cylinder_segment_2d(
        1.74, 2.46, np.pi * 5 / 6, n_ele_u=2, n_ele_v=3
    )

    # Create mesh
    mesh = Mesh()

    # Create patch set
    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        surf_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=element_description,
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_flat_plate_2d(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a two dimensional flat plate."""

    # Create the surface of a flat plate
    surf_obj = create_nurbs_flat_plate_2d(0.75, 0.91, n_ele_u=2, n_ele_v=5)

    # Create mesh
    mesh = Mesh()

    # Add material
    mat = get_default_test_solid_material(material_type="2d_shell")

    # Create patch set
    element_description = get_default_test_solid_element_description(
        element_type="2d_shell"
    )
    patch_set = add_geomdl_nurbs_to_mesh(
        mesh, surf_obj, material=mat, data=element_description
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_flat_plate_2d_splinepy(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a two dimensional flat plate with splinepy."""

    # Create the surface of a flat plate
    n_ele_u = 2
    n_ele_v = 5
    surf_obj = splinepy.helpme.create.box(0.75, 0.91).nurbs
    surf_obj.elevate_degrees([0, 1])
    surf_obj.insert_knots(0, np.linspace(0, 1, n_ele_u + 1))
    surf_obj.insert_knots(1, np.linspace(0, 1, n_ele_v + 1))
    ensure_3d_splinepy_object(surf_obj)
    translate_splinepy(surf_obj, -0.5 * np.array([0.75, 0.91, 0]))

    # Create the shell mesh
    mesh = Mesh()
    mat = get_default_test_solid_material(material_type="2d_shell")
    element_description = get_default_test_solid_element_description(
        element_type="2d_shell"
    )
    patch_set = add_splinepy_nurbs_to_mesh(
        mesh, surf_obj, material=mat, data=element_description
    )
    mesh.add(patch_set)
    assert_results_close(
        get_corresponding_reference_file_path(
            reference_file_base_name="test_integration_mesh_creation_functions_nurbs_flat_plate_2d"
        ),
        mesh,
    )


def test_integration_mesh_creation_functions_nurbs_flat_plate_2d_splinepy_copy(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that a mesh created from a splinepy NURBS can be copied."""

    # Create a flat plate
    surf_obj = splinepy.helpme.create.box(0.5, 1.0).nurbs
    surf_obj.elevate_degrees([0, 1])
    ensure_3d_splinepy_object(surf_obj)

    # Create mesh
    mesh = Mesh()
    mat = get_default_test_solid_material(material_type="2d_shell")
    element_description = get_default_test_solid_element_description(
        element_type="2d_shell"
    )
    add_splinepy_nurbs_to_mesh(mesh, surf_obj, material=mat, data=element_description)

    mesh_copy = mesh.copy()
    mesh_copy.translate([3, 0, 0])
    mesh.add(mesh_copy)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_brick(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a brick."""

    # Create a brick
    vol_obj = create_nurbs_brick(1.5, 3.0, 2.4, n_ele_u=2, n_ele_v=3, n_ele_w=4)

    # Create mesh
    mesh = Mesh()

    # Create patch set
    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        vol_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=get_default_test_solid_element_description(element_type="3d_solid"),
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_brick_splinepy(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a brick with splinepy."""

    # Create a brick
    n_el_dim = [2, 3, 4]
    box_dimensions = [1.5, 3.0, 2.4]
    vol_obj = splinepy.helpme.create.box(*box_dimensions).nurbs
    vol_obj.elevate_degrees([0, 1, 2])
    for i_dim, n_el in enumerate(n_el_dim):
        vol_obj.insert_knots(i_dim, np.linspace(0, 1, n_el + 1))

    control_points_3d = np.zeros([len(vol_obj.control_points), 3])
    control_points_3d[:, :3] = vol_obj.control_points
    vol_obj.control_points = control_points_3d - 0.5 * np.array(box_dimensions)

    # Create mesh
    mesh = Mesh()

    # Create patch set
    patch_set = add_splinepy_nurbs_to_mesh(
        mesh,
        vol_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=get_default_test_solid_element_description(element_type="3d_solid"),
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(
        get_corresponding_reference_file_path(
            reference_file_base_name="test_integration_mesh_creation_functions_nurbs_brick"
        ),
        mesh,
    )


def test_integration_mesh_creation_functions_nurbs_rotation_nurbs_surface(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the rotation of a NURBS mesh."""

    # Create the surface
    surf_obj = create_nurbs_hollow_cylinder_segment_2d(
        1.74, 2.46, np.pi * 3 / 4, n_ele_u=5, n_ele_v=2
    )

    # Create mesh
    mesh = Mesh()

    # Create patch set
    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        surf_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=element_description,
    )

    mesh.add(patch_set)

    mesh.rotate(Rotation([1, 2, 3], np.pi * 7 / 6))

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_translate_nurbs_surface(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the translation of a NURBS surface mesh."""

    # Create the surface
    surf_obj = create_nurbs_flat_plate_2d(0.87, 1.35, n_ele_u=2, n_ele_v=3)

    # Create mesh
    mesh = Mesh()

    # Create patch set

    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        surf_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=element_description,
    )

    mesh.add(patch_set)

    mesh.translate([-1.6, -2.3, 3.7])

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_cylindrical_shell_sector(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a 3-dimensional cylindrical shell sector."""

    # Create the surface of a quarter of a hollow cylinder
    surf_obj = create_nurbs_cylindrical_shell_sector(
        2.3, np.pi / 3, 1.7, n_ele_u=3, n_ele_v=5
    )

    # Create mesh
    mesh = Mesh()

    # Create patch set
    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        surf_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=get_default_test_solid_element_description(element_type="2d_solid"),
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_couple_nurbs_meshes(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the coupling of NURBS surface meshes."""

    # Create mesh
    mesh = Mesh()

    # Create the first surface object
    surf_obj_1 = create_nurbs_hollow_cylinder_segment_2d(
        0.65, 1.46, np.pi * 2 / 3, n_ele_u=3, n_ele_v=2
    )

    # Create first patch set
    mat = get_default_test_solid_material(material_type="st_venant_kirchhoff")

    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    patch_set_1 = add_geomdl_nurbs_to_mesh(
        mesh, surf_obj_1, material=mat, data=element_description
    )

    mesh.add(patch_set_1)

    mesh.rotate(Rotation([0, 0, 1], np.pi / 3))

    # Create the second surface object
    surf_obj_2 = create_nurbs_hollow_cylinder_segment_2d(
        0.65, 1.46, np.pi / 3, n_ele_u=3, n_ele_v=2
    )

    patch_set_2 = add_geomdl_nurbs_to_mesh(
        mesh, surf_obj_2, material=mat, data=element_description
    )

    mesh.add(patch_set_2)

    mesh.couple_nodes(reuse_matching_nodes=True)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_sphere_surface(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creating of the base patch of the surface of a sphere."""

    # Create mesh
    mesh = Mesh()

    # Create the base of a sphere
    surf_obj = create_nurbs_sphere_surface(1, n_ele_u=3, n_ele_v=2)

    # Create first patch set
    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        surf_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=element_description,
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_string_types(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creating of a NURBS with strings for the element and material
    definition."""

    # Create mesh
    mesh = Mesh()

    # Create the base of a sphere
    surf_obj = create_nurbs_flat_plate_2d(1, 3, n_ele_u=3, n_ele_v=2)

    # Create first patch set
    patch_set = add_geomdl_nurbs_to_mesh(
        mesh,
        surf_obj,
        material=get_default_test_solid_material(material_type="st_venant_kirchhoff"),
        data=get_default_test_solid_element_description(element_type="2d_solid"),
    )

    mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_hemisphere_surface(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of the surface of a hemisphere."""

    # Create mesh
    mesh = Mesh()

    # Create the base of a sphere
    surfs = create_nurbs_hemisphere_surface(2.5, n_ele_uv=2)

    # Create first patch set
    mat = get_default_test_solid_material(material_type="st_venant_kirchhoff")
    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    # Add the patch sets of every surface section of the hemisphere to the input file
    for surf in surfs:
        patch_set = add_geomdl_nurbs_to_mesh(
            mesh,
            surf,
            material=mat,
            data=element_description,
        )

        mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_torus_surface(
    get_default_test_solid_element_description,
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test the creation of a torus."""

    # Create mesh
    mesh = Mesh()

    # Create the surface of a torus
    surfs = create_nurbs_torus_surface(1, 0.5, n_ele_u=2, n_ele_v=3)

    # Define element description
    mat = get_default_test_solid_material(material_type="st_venant_kirchhoff")
    element_description = get_default_test_solid_element_description(
        element_type="2d_solid"
    )

    # Add the patch sets of every surface section of the torus to the input file
    for surf in surfs:
        patch_set = add_geomdl_nurbs_to_mesh(
            mesh,
            surf,
            material=mat,
            data=element_description,
        )

        mesh.add(patch_set)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)


def test_integration_mesh_creation_functions_nurbs_empty_knot_spans(
    get_default_test_solid_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Test that NURBS patches with empty knot spans are handled correctly."""

    # Create the pipe geometry with splinepy
    disk = splinepy.helpme.create.disk(
        outer_radius=2.3, inner_radius=1.7, angle=360, n_knot_spans=1
    )
    pipe = disk.create.extruded(extrusion_vector=[0.0, 0.0, 2.0])
    pipe.elevate_degrees([0, 2])
    pipe.uniform_refine(2, 2)

    # Create mesh
    mesh = Mesh()
    mat = get_default_test_solid_material(material_type="st_venant_kirchhoff")
    patch_set = add_splinepy_nurbs_to_mesh(mesh, pipe, material=mat)
    mesh.add(patch_set)
    mesh.couple_nodes(reuse_matching_nodes=True)

    # Compare with the reference file
    assert_results_close(get_corresponding_reference_file_path(), mesh)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to unittest the functionality of the geometry sets."""

from typing import Callable

import pytest

from beamme.core.conf import bme
from beamme.core.element_beam import Beam, Beam3
from beamme.core.geometry_set import GeometrySet, GeometrySetNodes
from beamme.core.mesh import Mesh
from beamme.core.node import Node, NodeCosserat
from beamme.core.rotation import Rotation
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


@pytest.fixture()
def assert_geometry_set_add_operator() -> Callable:
    """Return a function to check the results in the geometry set operator
    tests."""

    def _compare_results(
        mesh_objects, combined_geometry, set_1_geometry, set_2_geometry
    ):
        """Compare the results."""

        # Check that the added geometry set contains the combined geometry
        assert len(combined_geometry) == 5
        assert combined_geometry[0] is mesh_objects[2]
        assert combined_geometry[1] is mesh_objects[3]
        assert combined_geometry[2] is mesh_objects[4]
        assert combined_geometry[3] is mesh_objects[0]
        assert combined_geometry[4] is mesh_objects[1]

        # Check that the original sets are not modified
        assert len(set_1_geometry) == 3
        assert set_1_geometry[0] is mesh_objects[0]
        assert set_1_geometry[1] is mesh_objects[1]
        assert set_1_geometry[2] is mesh_objects[2]
        assert len(set_2_geometry) == 3
        assert set_2_geometry[0] is mesh_objects[2]
        assert set_2_geometry[1] is mesh_objects[3]
        assert set_2_geometry[2] is mesh_objects[4]

    return _compare_results


@pytest.mark.parametrize(
    ("mesh_object", "mesh_object_args"),
    [
        (Node, [[1, 2, 3]]),
        (Beam, []),
    ],
)
def test_beamme_core_geometry_set_add_operator(
    mesh_object, mesh_object_args, assert_geometry_set_add_operator
):
    """Test that geometry sets can be added to each other.

    We test this once with a point geometry set based on nodes and once
    with a line geometry set based on beam elements.
    """

    mesh_objects = [mesh_object(*mesh_object_args) for _ in range(5)]
    set_1 = GeometrySet(mesh_objects[:3])
    set_2 = GeometrySet(mesh_objects[2:])
    combined_set = set_2 + set_1
    combined_geometry = combined_set.get_geometry_objects()

    assert_geometry_set_add_operator(
        mesh_objects,
        combined_geometry,
        set_1.get_geometry_objects(),
        set_2.get_geometry_objects(),
    )


@pytest.mark.parametrize("geometry_type", [bme.geo.point, bme.geo.line])
def test_beamme_core_geometry_set_nodes_add_operator(
    geometry_type, assert_geometry_set_add_operator
):
    """Test that node based geometry sets can be added to each other."""

    mesh_objects = [Node([1, 2, 3]) for _ in range(5)]
    set_1 = GeometrySetNodes(geometry_type, nodes=mesh_objects[:3])
    set_2 = GeometrySetNodes(geometry_type, nodes=mesh_objects[2:])
    combined_set = set_2 + set_1
    combined_geometry = combined_set.get_all_nodes()

    assert_geometry_set_add_operator(
        mesh_objects, combined_geometry, set_1.get_all_nodes(), set_2.get_all_nodes()
    )


def test_beamme_core_geometry_set_add():
    """Test functionality of the GeometrySet add method."""

    mesh = Mesh()
    for i in range(6):
        mesh.add(NodeCosserat([i, 2 * i, 3 * i], Rotation()))

    set_1 = GeometrySetNodes(
        bme.geo.point, [mesh.nodes[0], mesh.nodes[1], mesh.nodes[2]]
    )
    set_2 = GeometrySetNodes(
        bme.geo.point, [mesh.nodes[2], mesh.nodes[3], mesh.nodes[4]]
    )
    set_12 = GeometrySetNodes(bme.geo.point)
    set_12.add(set_1)
    set_12.add(set_2)
    set_3 = GeometrySet(set_1.get_points())

    mesh.add(set_1, set_2, set_12, set_3)

    # Check the resulting sets
    unique_sets = mesh.get_unique_geometry_sets()
    for key, value in unique_sets.items():
        if key is bme.geo.point:
            assert len(value) == 4
        else:
            assert len(value) == 0

    results = [
        {"len": 3, "indices": [0, 1, 2]},
        {"len": 3, "indices": [2, 3, 4]},
        {"len": 5, "indices": [0, 1, 2, 3, 4]},
        {"len": 3, "indices": [0, 1, 2]},
    ]
    for i_set, result_dict in enumerate(results):
        point_set = unique_sets[bme.geo.point][i_set]
        nodes = point_set.get_all_nodes()
        assert len(nodes) == result_dict["len"]
        for i_node, node_index in enumerate(result_dict["indices"]):
            assert nodes[i_node] is mesh.nodes[node_index]


def test_beamme_core_geometry_set_unique_ordering_of_get_all_nodes_for_line_condition(
    get_default_test_beam_material,
):
    """This test ensures that the ordering of the nodes returned from the
    function get_all_nodes is unique for line sets."""

    # set up a beam mesh with material
    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="base")
    beam_set = create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [2, 0, 0], n_el=10)

    # check the nodes in the line set
    nodes_set = beam_set["line"].get_all_nodes()
    assert len(nodes_set) == 21
    mesh_node_indices = range(21)
    for i_node_set, i_node_mesh in enumerate(mesh_node_indices):
        assert nodes_set[i_node_set] is mesh.nodes[i_node_mesh]


def test_beamme_core_geometry_set_get_geometry_objects(get_default_test_beam_material):
    """Test if the geometry set returns the objects(elements) in the correct
    order."""

    # Initialize material and mesh
    mat = get_default_test_beam_material(material_type="base")
    mesh = Mesh()

    # number of elements
    n_el = 5

    # Create a simple beam.
    geometry = create_beam_mesh_line(mesh, Beam3, mat, [0, 0, 0], [2, 0, 0], n_el=n_el)

    # Get all elements from the geometry set.
    elements_of_geometry = geometry["line"].get_geometry_objects()

    # Check number of elements.
    assert len(elements_of_geometry) == n_el

    # Check if the order of the elements from the geometry set is the same as for the mesh.
    for i_element, element in enumerate(elements_of_geometry):
        assert element == mesh.elements[i_element]
